import re
from lxml.etree import Element, fromstring, tostring

import lupa
from copy import deepcopy
from typing import Any, Callable
from textwrap import indent, dedent

from celadon import Widget, load_rules, Page

from .lua import lua, LuaTable, WIDGET_TYPES
from .callbacks import parse_callback

STYLE_TEMPLATE = """\
{query}:
{indented_content}\
"""

LUA_SCRIPT_BEGIN = """\
do table.insert(stack, _ENV)
    _ENV = initScope(_ENV)
    env_id = {script_id}

    -- USER CODE BEGIN
"""

LUA_SCRIPT_END = """
    -- USER CODE END

envs[{script_id}] = _ENV
end _ENV = table.remove(stack)
if _children then table.insert(_children, envs[{script_id}]) end

"""

EVENT_PREFIXES = ("on", "pre")

# TODO: This breaks `width: shrink` for text
def lua_formatted_get_content(scope: dict[str, Any]) -> Callable[[Widget], list[str]]:
    """Returns a `get_content` method that formats Lua variables.

    You can use any variable available in the current scope using a $ prefix, like
    `$count`.
    """

    def _get_content(self: Widget) -> list[str]:
        lines = []

        for line in self.__class__.get_content(self):
            in_var = False
            parts = []
            word = ""

            for word in re.split(r"([^a-zA-Z0-9_\.])", line):
                if word == "$":
                    in_var = True
                    continue

                if in_var:
                    in_var = False

                    value = scope[word]

                    if value is None:
                        raise ValueError(
                            f"unknown variable '{word}' for {self.as_query()}"
                        )

                    word = str(value)

                parts.append(word)

            lines.append("".join(parts))

        return lines

    return _get_content


def parse_rules(text: str, query: str | None = None) -> dict[str, Any]:
    """Parses a block of YAML rules into a dictionary."""

    if query is None:
        style = dedent(text)
    else:
        style = STYLE_TEMPLATE.format(
            query=query, indented_content=indent(dedent(text), 4 * " ")
        )

    return load_rules(style)


def _extract_script(
    node: Element, node_to_id: dict[Element, int], outer: bool = False, level: int = 0
) -> str:
    """Recursively extracts scripts starting from the given node."""

    code = ""

    if outer:
        code += "_ENV = sandbox.envs[0]\n\n"

    code += indent(
        LUA_SCRIPT_BEGIN.format(script_id=node_to_id[node]),
        level * 4 * " ",
    )

    for child in node:
        if child.tag == "style":
            continue

        if child.tag == "script":
            code += indent(dedent(child.text), (level + 1) * 4 * " ")
            continue

        code += _extract_script(child, node_to_id, level=level + 1)

    code += indent(LUA_SCRIPT_END.format(script_id=node_to_id[node]), level * 4 * " ")

    return code


def _get_pairs(table: LuaTable) -> list[str]:
    """Iterates through the `__pairs` of a Lua table."""

    iterator, state, first_key = lua.globals().pairs(table)

    while True:
        item = iterator(state, first_key)

        if item is None:
            break

        key, value = item

        yield (key, value)
        first_key = key


# TODO: Technically rules is more like a `dict[str, dict[str, <something>]]`!
def parse_widget(
    node: Element,
    components: dict[str, tuple[dict[str, Any], str]],
    parse_script: bool = True,
    result: dict[str, list[Widget, Element]] | None = None,
) -> tuple[Widget, dict[str, Any]]:
    """Parses a widget, its scripts & its styling from an XML node."""

    result = result or {}

    init: dict[str, str | tuple[str, ...] | list[Callable[[Widget], bool]]] = {}

    if node.tag in components:
        params, replacement = components[node.tag]
        replacement = deepcopy(replacement)

        script = replacement.find("script")

        if script is None:
            script = Element("script")
            script.text = " "

            node.append(script)

        slot = replacement.find("_slot")

        if slot is not None:
            slot_idx = [*replacement].index(slot)
            replacement.remove(slot)

            content = [*node]

            for i, child in enumerate(content):
                replacement[max(slot_idx - 1 + i, 0)].addnext(child)

        parent = node.getparent()
        idx = [*parent].index(node)

        text = tostring(replacement).decode()

        for key, value in params.items():
            text = text.replace(f"${key}", node.get(key, default=value))

        replacement = fromstring(text.encode())

        node = replacement
        parent[idx] = replacement

    for key, value in node.attrib.items():
        if key == "groups":
            init["groups"] = tuple(value.split(" "))
            continue

        if key.startswith(EVENT_PREFIXES):
            key = key.replace("-", "_")

            if value.startswith(":"):
                init[key] = [parse_callback(value)]

            else:
                script_node = node.find("script")

                if script_node is None:
                    script_node = Element("script")
                    script_node.text = ""

                script_node.text += f"function {key}() {value} end\n\n"
                node.append(script_node)

            continue

        key = key.replace("-", "_")

        if value.isdigit():
            value = int(value)

        elif value.lstrip("-+").replace(".", "", 1).isdigit():
            value = float(value)

        init[key] = value

    text = node.text

    if text is None or text.strip() == "":
        text = ""
        skipped = 0
        total = 0

        for total, child in enumerate(node):
            if child.tail is None:
                skipped += 1
                continue

            text = text + child.tail

        if skipped == total + 1:
            text = None

    cls = WIDGET_TYPES[node.tag]

    if text is not None and text.strip() != "":
        # The init args aren't strongly typed.
        widget = cls(dedent(text).strip("\n"), **init)  # type: ignore
    else:
        widget = cls(**init)  # type: ignore

    query = widget.as_query()
    scope = None

    rules: dict[str, Any] = {}

    script_id = id(widget)
    result[script_id] = widget, node

    for child in node:
        if child.tag == "style":
            rules.update(**parse_rules(child.text or "", query))
            continue

        if child.tag == "script":
            continue

        parsed, parsed_rules = parse_widget(
            child, components, parse_script=False, result=result
        )
        rules.update(**parsed_rules)
        widget += parsed  # type: ignore

        if child.tag in WIDGET_TYPES:
            result[id(parsed)] = parsed, child

    if parse_script:
        code = _extract_script(
            node, {node: s_id for s_id, [_, node] in result.items()}, outer=True
        )

    if not parse_script:
        return widget, rules

    sandbox = lua.eval("sandbox")
    envs = lua.eval("sandbox.envs")
    setfenv = lua.eval("builtins.setfenv")

    try:
        lua.execute(code)
    except lupa.LuaSyntaxError as exc:
        # TODO: This could alert() instead and abort exec
        raise exc

    for s_id, [widget, _] in reversed(result.items()):
        env = envs[s_id]

        if env is None:
            continue

        env.self = widget

        for key, value in _get_pairs(env):
            if sandbox[key] is not None:
                continue

            if callable(value):
                if not env.hasOwn(key):
                    continue

                value = setfenv(value, env)

            if key == "init":
                value()
                continue

            if isinstance(key, str) and key.startswith(EVENT_PREFIXES):
                event = getattr(widget, key, None)

                if event is None:
                    raise ValueError(f"invalid event handler {key!r}")

                assert callable(value)

                event += value

        # Set formatted get content for the widget
        get_content = lua_formatted_get_content(env)
        widget.get_content = get_content.__get__(widget, widget.__class__)  # type: ignore

    return widget, rules


def _register_component(
    node: Element, components: dict[str, str], namespace: str | None = None
) -> None:
    name = None
    params = {}

    for key, value in node.attrib.items():
        if key == "name":
            name = value
            continue

        params[key] = value

    if name is None:
        raise ValueError("components must have a name.")

    if namespace is not None:
        name = namespace + "." + name

    components[name] = params, node[0]


def parse_page(node: Element, components: dict[str, str]) -> Page:
    """Parses a page, its scripts & its children from XML node."""

    page_node = node.find("page")

    if page_node is None:
        raise ValueError("no <page /> node found.")

    # Init args aren't strongly typed
    page = Page(**page_node.attrib)  # type: ignore

    for child in page_node:
        if child.tag == "complib":
            namespace = child.get("namespace")

            for inner in child:
                _register_component(inner, components, namespace)

            continue

        if child.tag == "component":
            _register_component(child, components)
            continue

        if child.tag in WIDGET_TYPES or child.tag in components:
            widget, rules = parse_widget(child, components)
            page += widget

            for selector, rule in rules.items():
                page.rule(selector, **rule)

        elif child.tag == "style":
            for selector, rule in parse_rules(child.text or "").items():
                page.rule(selector, **rule)

        else:
            raise ValueError(child.tag)

    return page



from functools import wraps
from typing import Any, Callable
from threading import Thread
from urllib.parse import urlparse

from lxml.etree import fromstring as ElementTree, Element
from xml.dom.minidom import parseString
from xml.parsers.expat import ExpatError

from celadon import Application, Page, Widget, Container
from zenith import zml_escape
from requests import Session

from .parsing import parse_widget, parse_page
from .callbacks import (
    HTTPMethod,
    Instruction,
    Verb,
    TreeMethod,
)
from .lua import lua, init_runtime


__all__ = ["HttpApplication"]


def threaded(func: Callable[..., None]) -> Callable[..., None]:
    """Returns a callable that runs the given function in a thread."""

    @wraps(func)
    def _inner(*args, **kwargs) -> None:
        Thread(target=func, args=args, kwargs=kwargs).start()

    return _inner


class HttpApplication(Application):
    """An application class for HTTP pages."""

    def __init__(self, domain: str, **app_args: Any) -> None:
        super().__init__(**app_args)

        init_runtime(lua, self)

        self._registered_components = {}
        self._page = Page()
        self._url = urlparse(domain)
        self._session = Session()
        self._session.headers = {
            "Accepts": "text/celx",
            "CELX_Request": "true",
        }
        self._current_instructions: list[list[Instruction]] = []

        def _clear_instructions(_: Page) -> bool:
            for instructions in self._current_instructions:
                instructions.clear()

            return True

        self.on_page_changed += _clear_instructions

        self.route(self._url.geturl())

    @property
    def session(self) -> Session:
        """Returns the current requests session."""

        return self._session

    def __getitem__(self, item: Any) -> Any:
        """Implement `__getitem__` for Lua attribute access."""

        return getattr(self, item)

    def _error(self, error: Exception) -> None:
        self.stop()

        self._raised = error

    def _prefix_endpoint(self, endpoint: str) -> str:
        """Prefixes hierarchy-only endpoints with the current url and its scheme."""

        if endpoint.startswith("/"):
            return self._url.scheme + "://" + self._url.netloc + endpoint

        return endpoint

    def _http(
        self,
        method: HTTPMethod,
        endpoint: str,
        data: dict[str, Any],
        handler: Callable[[Element], None],
    ) -> Thread:
        endpoint = self._prefix_endpoint(endpoint)

        if method is HTTPMethod.GET:
            request_data = {"params": data}
        else:
            request_data = {"data": data}

        if not isinstance(method, HTTPMethod):
            self._error(TypeError(f"Invalid method {method!r}."))

        request = getattr(self._session, method.value.lower())

        def _execute() -> None:
            resp = request(endpoint, **request_data)

            if not 200 <= resp.status_code < 300:
                self.stop()
                resp.raise_for_status()

            url = urlparse(endpoint)

            if url.netloc != self._url.netloc:
                self._url = url

            # Wrap invalid XML as text
            # TODO: Treat response differently based on Content-Type
            xml = resp.text

            try:
                _ = parseString(xml)
            except ExpatError:
                xml = zml_escape(xml)
                xml = f"<text>{xml}</text>"

            tree = ElementTree(xml)

            for sourceable in ["style", "script", "complib"]:
                for node in tree.findall(f".//{sourceable}[@src]"):
                    resp = self._session.get(self._prefix_endpoint(node.attrib["src"]))

                    if not 200 <= resp.status_code < 300:
                        self.stop()
                        resp.raise_for_status()

                    if sourceable == "complib":
                        sourced = ElementTree(resp.text)
                        for child in sourced:
                            node.append(child)

                        for key, value in sourced.attrib.items():
                            node.attrib[key] = value

                    else:
                        node.text = resp.text

                    del node.attrib["src"]

            return handler(tree)

        thread = Thread(target=_execute)
        thread.start()

        return thread

    def _xml_page_route(self, node: Element) -> None:
        """Routes to a page loaded from the given XML."""

        try:
            page = parse_page(node, self._registered_components)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self._error(exc)
            return

        page.route_name = self._url.path

        # TODO: We don't have to request the page every time we go to it
        self.append(page)

        self._page = page
        self._mouse_target = self._page[0]
        self.on_page_changed(page)

        if page.route_name == "/":
            self._terminal.set_title(self.title)

        else:
            self._terminal.set_title(page.title)

        self.apply_rules()

    @threaded
    def run_instructions(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        self, instructions: list[Instruction], caller: Widget
    ) -> None:
        """Runs through a list of instructions."""

        result = None

        def _set_result(xml: Element) -> None:
            nonlocal result

            # Drill down to find the first widget, pass that on instead.
            if xml.tag == "celx":
                for node in xml.findall("./page//"):
                    if node.tag not in ["style", "script"]:
                        xml = node
                        break

                else:
                    self._error(ValueError("no widget in response"))
                    return

            result, rules = parse_widget(xml, self._registered_components)

            if self.page is None:
                return

            # TODO: There might be cases where we don't want to apply styles immediately,
            #       like when a future "DELETE" instruction is added.
            for selector, rule in rules.items():
                self.page.rule(selector, **rule)

        self._current_instructions.append(instructions)

        try:  # pylint: disable=too-many-nested-blocks
            for instr in instructions:
                if instr.verb.value in HTTPMethod.__members__:
                    endpoint, container = instr.args
                    assert endpoint is not None

                    body: Widget | Page | None = caller.parent

                    if container is not None:
                        body = self.find(container)

                        if body is None:
                            raise ValueError(f"nothing matched selector {container!r}")

                    if not isinstance(body, Widget):
                        raise ValueError(f"request body {body!r} is not serializable")

                    content = body.serialize()

                    self._http(
                        HTTPMethod(instr.verb.value), endpoint, content, _set_result
                    ).join()

                    continue

                if instr.verb.value in TreeMethod.__members__:
                    if result is None:
                        raise ValueError("no result to update tree with")

                    selector, modifier = instr.args
                    assert selector is not None

                    target = self.find(selector)

                    if target is None:
                        raise ValueError(f"nothing matched selector {selector!r}")

                    if not isinstance(target, Container):
                        raise ValueError(
                            f"cannot modify tree of non-container {target!r}"
                        )

                    if instr.verb is Verb.SWAP:
                        offsets = {"before": -1, None: 0, "after": 1}

                        if modifier == "IN":
                            target.update_children([result])

                        elif modifier in offsets:
                            if not isinstance(target.parent, Container):
                                raise ValueError(
                                    "cannot modify tree of non-container parent of"
                                    + repr(target)
                                )

                            target.parent.replace(
                                target, result, offset=offsets[modifier]
                            )

                        else:
                            raise ValueError(
                                f"unknown modifier {modifier!r} for verb {instr.verb!r}"
                            )

                    elif instr.verb is Verb.INSERT:
                        if modifier == "IN":
                            target.insert(0, result)

                        else:
                            if not isinstance(target.parent, Container):
                                raise ValueError(
                                    "cannot modify tree of non-container parent of"
                                    + repr(target)
                                )

                            index = target.parent.children.index(target)
                            offsets = {"before": index, "after": index + 1}

                            if modifier not in offsets:
                                raise ValueError(
                                    f"unknown modifier {modifier!r}"
                                    + f" for verb {instr.verb!r}"
                                )

                            target.parent.insert(offsets[modifier], result)

                    elif instr.verb is Verb.APPEND:
                        if modifier != "IN":
                            raise ValueError(
                                f"unknown modifier {modifier!r} for verb {instr.verb!r}"
                            )

                        target.append(result)

                    # TODO: This is hacky as hell, but we need it for widgets to load in
                    #       .styles
                    parent = result.parent
                    self._init_widget(result)
                    result.parent = parent

                    continue

                if instr.verb is Verb.SELECT:
                    if result is None:
                        raise ValueError("no result to select from")

                    if not isinstance(result, Container):
                        raise ValueError(
                            f"cannot select from non container ({result!r})"
                        )

                    assert instr.args[0] is not None

                    result = self.find(instr.args[0], scope=result)
                    continue

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self._error(exc)
            return

        self._current_instructions.remove(instructions)

    def route(self, destination: str) -> None:
        """Routes to the given URL."""

        destination = self._prefix_endpoint(destination)

        self._http(HTTPMethod.GET, destination, {}, self._xml_page_route)



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Type, TypeVar
from textwrap import dedent

from lupa import LuaRuntime  # type: ignore # pylint: disable=no-name-in-module
from celadon import Widget, widgets
from zenith import zml_alias, zml_macro, MacroType, zml_escape, zml_expand_aliases

from .callbacks import parse_callback

if TYPE_CHECKING:
    from .application import HttpApplication

WIDGET_TYPES = {
    key.lower(): value
    for key, value in vars(widgets).items()
    if isinstance(value, type) and issubclass(value, Widget)
}

LuaTable = TypeVar("LuaTable")

LUA_SCOPE_SETUP = """
builtins = {
    ipairs = ipairs,
    next = next,
    pairs = pairs,
    pcall = pcall,
    tonumber = tonumber,
    tostring = tostring,
    type = type,
    coroutine = { create = coroutine.create, resume = coroutine.resume,
        running = coroutine.running, status = coroutine.status,
        wrap = coroutine.wrap },
    string = { byte = string.byte, char = string.char, find = string.find,
        format = string.format, gmatch = string.gmatch, gsub = string.gsub,
        len = string.len, lower = string.lower, match = string.match,
        rep = string.rep, reverse = string.reverse, sub = string.sub,
        upper = string.upper },
    table = { insert = table.insert, maxn = table.maxn, remove = table.remove,
        sort = table.sort, unpack = table.unpack },
    math = { abs = math.abs, acos = math.acos, asin = math.asin,
        atan = math.atan, atan2 = math.atan2, ceil = math.ceil, cos = math.cos,
        cosh = math.cosh, deg = math.deg, exp = math.exp, floor = math.floor,
        fmod = math.fmod, frexp = math.frexp, huge = math.huge,
        ldexp = math.ldexp, log = math.log, log10 = math.log10, max = math.max,
        min = math.min, modf = math.modf, pi = math.pi, pow = math.pow,
        rad = math.rad, random = math.random, sin = math.sin, sinh = math.sinh,
        sqrt = math.sqrt, tan = math.tan, tanh = math.tanh },
    os = { clock = os.clock, difftime = os.difftime, time = os.time },
    copy = copy, 
    setmetatable = setmetatable,
    print = print,
    error = error,
    debug = { getupvalue = debug.getupvalue, upvaluejoin = debug.upvaluejoin },
    dump_keys = function(t)
        local result = ""

        for k, _ in pairs(t) do
            result = result .. tostring(k) .. ", "
        end

        return result
    end,
    dump_values = function(t)
        local result = ""

        for _, v in pairs(t) do
            result = result .. tostring(v) .. ", "
        end

        return result
    end,
    setfenv = function(fn, env)
        local i = 1
        while true do
            local name = debug.getupvalue(fn, i)
            if name == "_ENV" then
                debug.upvaluejoin(fn, i, (function()
                    return env
                end), 1)
                break
            elseif not name then
                break
            end

            i = i + 1
        end

        return fn
    end,
}

sandbox = {
    builtins = builtins,
    app = nil,
    stack = {},
    envs = {},

    alert = function(text)
        local text = tostring(text)

        app.pin(w.Dialogue{
            w.Text{text, group="body"},
            w.Row{
                w.Button{"Close", on_submit={
                    function() app.unpin_last() end
                }},
                group="input",
            }
        })
    end,

    confirm = function(title, body, callback)
        app.pin(w.Dialogue{
            w.Text{title, group="title"},
            w.Text{body, group="body"},
            w.Row{
                w.Button{
                    "Confirm",
                    on_submit={function()
                        app.unpin_last()
                        callback(true)
                    end}
                },
                w.Button{
                    "Deny",
                    on_submit={function()
                        app.unpin_last()
                        callback(false)
                    end}
                },
                group="input"
            }
        })
    end,

    prompt = function(title, body, callback)
        app.pin(w.Dialogue{
            w.Text{tostring(title)},
            w.Tower{group="body", table.unpack(body)},
            w.Row{
                w.Button{"Submit", on_submit={function()
                    callback(app.unpin_last())
                end}},
                group="input"
            }
        })
    end,

    find = function(selector, multiple)
        if multiple then
            return table.from_py(app.find_all(selector))
        end

        return app.find(selector)
    end,

    initScope = function(hiddenScope)
        local innerScope = {}

        local _listeners = {}
        local _children = {}

        return setmetatable({
            on_change = function(field, callback)
                if _listeners[field] == nil then
                    _listeners[field] = {}
                end

                table.insert(_listeners[field], callback)
            end,

            hasOwn = function(k)
                return innerScope[k] ~= nil
            end,

            builtins = builtins,

            _children = _children,
            _listeners = _listeners,
        }, {
            __index = function(t, k)
                local value = innerScope[k]

                if value == nil then
                    value = hiddenScope[k]
                end

                if value == nil then
                    value = builtins[k]
                end

                if builtins.type(value) == "function" then
                    builtins.setfenv(value, t)
                end

                return value
            end,

            __pairs = function(t)
                local merged = {{}}

                for k, v in pairs(hiddenScope) do
                    merged[k] = v
                end

                for k, v in pairs(innerScope) do
                    merged[k] = v
                end

                return pairs(merged)
            end,

            __newindex = function(t, k, v)
                local current = hiddenScope[k]

                if current == nil or builtins.type(current) == "function" then
                    innerScope[k] = v
                else
                    hiddenScope[k] = v
                end

                if current ~= v and _listeners[k] ~= nil then
                    for _, callback in ipairs(_listeners[k]) do
                        builtins.setfenv(callback, t)(v)
                    end

                    for _, subenv in ipairs(_children) do
                        local listeners = subenv._listeners[k]

                        if listeners ~= nil then
                            for _, callback in pairs(listeners) do
                                builtins.setfenv(callback, subenv)(v)
                            end
                        end
                    end
                end
            end,
        })
    end
}
"""


def _attr_filter(_, attr, __):
    """Removes access to sunder and dunder attributes in Lua code."""

    if not isinstance(attr, str):
        return attr

    if attr.startswith("_"):
        raise AttributeError("access denied")

    return attr


class LuaStyleWrapper:
    """Wraps a widget's style object for nice Lua syntax.

    ```
    styles.content = 'red'
    styles(some_widget).content = 'blue'
    ```
    """

    def __init__(self, widget: Widget) -> None:
        self._widget = widget

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr == "_widget":
            super().__setattr__(attr, value)

            return

        self._widget.app.rule(
            self._widget.as_query(), score=9999, **{f"{attr}_style": value}
        )

    def __call__(self, widget: Widget) -> LuaStyleWrapper:
        return self.__class__(widget)


def _multi_find(
    app: "HttpApplication", selector: str, multiple: bool = False
) -> Widget | list[Widget] | None:
    """Finds widgets from the application."""

    """
    """

    if multiple:
        return lua.table_from([*app.find_all(selector)])

    return app.find(selector)


def _zml_define(name: str, macro: MacroType) -> None:
    """Defines a macro in Lua."""

    def _inner(*args) -> str:
        return macro(*args)

    _inner.__name__ = name.replace(" ", "_")
    zml_macro(_inner)


def _chocl(descriptor: str) -> Callable[[Widget], bool]:
    """Provides Lua access to the celx hypermedia-oriented callback language."""

    return parse_callback(descriptor)


def _remove_and_callback_with(
    dialogue: widgets.Dialogue, callback: Callable, value: Any
) -> bool:
    dialogue.remove_from_parent()
    callback(value)

    return False


def _widget_factory(envs: LuaTable, typ: Type[Widget]) -> Callable[[LuaTable], Widget]:
    """Lets Lua instantiate widgets.

    ```
    Button{"label", on_submit={function() alert("hey") end }}
    ```
    """

    setfenv = lua.globals().builtins.setfenv
    getenv = lua.globals().sandbox.env

    widget = None

    def _create(options: LuaTable) -> Widget:
        nonlocal widget

        args = []
        kwargs = {}

        for key, value in options.items():
            if isinstance(key, int):
                args.append(value)
                continue

            if key == "groups":
                value = tuple(value.values())

            # Likely a Lua table
            elif hasattr(value, "values"):
                if all(isinstance(val, int) for val in list(value)):
                    value = [*value.values()]
                else:
                    value = dict(value.items())

            if callable(value):
                value = _scope_func(value)

            kwargs[key] = value

        widget = typ(*args, **kwargs)
        return widget

    return _create


def _env_getter(envs: LuaTable) -> Callable[[Widget], LuaTable]:
    def _inner(widget: Widget) -> LuaTable:
        return envs[id(widget)]

    return _inner


def init_runtime(runtime: LuaRuntime, app: "HttpApplication") -> None:
    """Sets up the global namespace for the given runtime."""

    runtime.execute(LUA_SCOPE_SETUP)

    sandbox = runtime.globals().sandbox
    runtime.globals().builtins.table.from_py = runtime.table_from

    sandbox.app = app
    sandbox.timeout = app.timeout

    sandbox.env = _env_getter(sandbox.envs)
    sandbox.styles = LuaStyleWrapper
    sandbox.chocl = parse_callback

    sandbox.zml = {
        "alias": zml_alias,
        "define": _zml_define,
        "escape": zml_escape,
        "expand_aliases": zml_expand_aliases,
    }
    sandbox.w = {
        key.title(): _widget_factory(sandbox.envs, value)
        for key, value in WIDGET_TYPES.items()
    }

    runtime.execute(
        dedent(
            """
        do table.insert(sandbox.stack, _ENV)
            scope = {}
            for k, v in pairs(sandbox) do
                scope[k] = v
            end

            sandbox.envs[0] = sandbox.initScope(scope)
        end
    """
        )
    )


class LoggedLuaRuntime(LuaRuntime):
    FILE = "debug.lua"

    def execute(self, code: str, **kwargs) -> Any:
        if self.FILE:
            with open(self.FILE, "a") as f:
                f.write(code + "\n")

        return super().execute(code, **kwargs)


lua = LoggedLuaRuntime(
    register_eval=False,
    register_builtins=False,
    unpack_returned_tuples=True,
    attribute_filter=_attr_filter,
)
